###########################################
#
# Example script to build a
# pinout diagram. Includes basic
# features and convenience classes.
#
###########################################

from pinout.core import Group, Image
from pinout.components.layout import Diagram_2Rows
from pinout.components.pinlabel import PinLabelGroup, PinLabel
from pinout.components.text import TextBlock
from pinout.components import leaderline as lline
from pinout.components.legend import Legend


# Import data for the diagram
import data

# Create a new diagram
# The Diagram_2Rows class provides 2 panels,
# 'panel_01' and 'panel_02', to insert components into.
diagram = Diagram_2Rows(1024, 576, 440, "diagram")

# Add a stylesheet
diagram.add_stylesheet("styles.css", embed=True)

# Create a group to hold the pinout-diagram components.
graphic = diagram.panel_01.add(Group(400, 42))

# Add and embed an image
hardware = graphic.add(Image("hardware.png", embed=True))

# Measure and record key locations with the hardware Image instance
hardware.add_coord("gpio0", 16, 100)
hardware.add_coord("gpio3", 65, 244)
hardware.add_coord("reset", 155, 244)
hardware.add_coord("vcc", 206, 100)
# Other (x,y) pairs can also be stored here
hardware.add_coord("pin_pitch_v", 0, 30)
hardware.add_coord("pin_pitch_h", 30, 0)

# Create a single pin label
graphic.add(
    PinLabel(
        content="RESET",
        x=hardware.coord("reset").x,
        y=hardware.coord("reset").y,
        tag="pwr",
        body={"x": 117, "y": 30},
        leaderline={"direction": "vh"},
    )
)

# Create pinlabels on the right header
graphic.add(
    PinLabelGroup(
        x=hardware.coord("vcc").x,
        y=hardware.coord("vcc").y,
        pin_pitch=hardware.coord("pin_pitch_v", raw=True),
        label_start=(60, 0),
        label_pitch=(0, 30),
        labels=data.right_header,
    )
)

# Create pinlabels on the left header
graphic.add(
    PinLabelGroup(
        x=hardware.coord("gpio0").x,
        y=hardware.coord("gpio0").y,
        pin_pitch=hardware.coord("pin_pitch_v", raw=True),
        label_start=(60, 0),
        label_pitch=(0, 30),
        scale=(-1, 1),
        labels=data.left_header,
    )
)

# Create pinlabels on the lower header
graphic.add(
    PinLabelGroup(
        x=hardware.coord("gpio3").x,
        y=hardware.coord("gpio3").y,
        scale=(-1, 1),
        pin_pitch=hardware.coord("pin_pitch_h", raw=True),
        label_start=(110, 30),
        label_pitch=(0, 30),
        labels=data.lower_header,
        leaderline=lline.Curved(direction="vh"),
    )
)

# Create a title and description text-blocks
title_block = diagram.panel_02.add(
    TextBlock(
        data.title,
        x=20,
        y=30,
        line_height=18,
        tag="panel title_block",
    )
)
diagram.panel_02.add(
    TextBlock(
        data.description,
        x=20,
        y=60,
        width=title_block.width,
        height=diagram.panel_02.height - title_block.height,
        line_height=18,
        tag="panel text_block",
    )
)

# Create a legend
legend = diagram.panel_02.add(
    Legend(
        data.legend,
        x=340,
        y=8,
        max_height=132,
    )
)

# Export the diagram via commandline:
# >>> py -m pinout.manager --export pinout_diagram.py diagram.svg



import math
from pinout import core


class Leaderline(core.Path):
    """Leaderline base object."""

    def __init__(self, direction="hh", **kwargs):
        self.direction = direction
        self.start = core.Coords(0, 0)
        self.end = core.Coords(0, 0)
        super().__init__(**kwargs)

    def end_points(self, origin, destination):
        """Locate origin and destination coordinates."""
        # origin and destination are components with bounding-boxes
        # direction is a 2 char code representing starting and ending directions
        # 'h' horizontal, 'v' vertical
        o_coords = origin.bounding_coords()
        d_coords = destination.bounding_coords()

        start = {
            "h": core.Coords(o_coords.x2, o_coords.y1 + origin.height / 2),
            "v": core.Coords(origin.x + (o_coords.x2 - o_coords.x1) / 2, o_coords.y2),
        }
        end = {
            "h": core.Coords(d_coords.x1, d_coords.y1 + destination.height / 2),
            "v": core.Coords(
                destination.x + (d_coords.x2 - d_coords.x1) / 2, d_coords.y1
            ),
        }
        self.start = start[self.direction[0]]
        self.end = end[self.direction[-1]]
        return (self.start, self.end)

    """
    def bounding_coords(self):
        return core.BoundingCoords(
            min(self.start.x, self.end.x),
            min(self.start.y, self.end.y),
            max(self.start.x, self.end.x),
            max(self.start.y, self.end.y),
        )
    """


class Curved(Leaderline):
    """Leaderline comprised of one or two curved corners."""

    def route(self, origin, destination):

        o_coords = origin.bounding_coords()
        d_coords = destination.bounding_coords()

        r = min(abs(origin.x - destination.x), abs(origin.y - destination.y)) / 3
        len = (d_coords.x1 - o_coords.x2) / 8
        ctl_h = (d_coords.x1 - o_coords.x2) / 2
        ctl_v = (d_coords.y1 - o_coords.y2) / 2

        start, end = self.end_points(origin, destination)

        if self.direction == "vh":
            # start vertical, end horizontal
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {start.x} {end.y - r}",
                    f"A {r} {r} 0 0 0 {start.x + r} {end.y}",
                    f"L {end.x} {end.y}",
                ]
            )
        elif self.direction == "hv":
            # start horizontal, end vertical
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {end.x - r} {start.y}",
                    f"A {r} {r} 0 0 1 {end.x} {start.y + r}",
                    f"L {end.x} {end.y}",
                ]
            )
        elif self.direction == "hh":
            # start horizontal, end horizontal
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {start.x + len} {start.y}",
                    f"C {start.x + ctl_h} {start.y} {end.x - ctl_h} {end.y} {end.x - len} {end.y}",
                    f"L {end.x} {end.y}",
                ]
            )
        elif self.direction == "vv":
            # start vertical, end vertical
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {start.x} {start.y + len}",
                    f"C {start.x} {start.y + ctl_v} {end.x} {end.y - ctl_v} {end.x} {end.y - len}",
                    f"L {end.x} {end.y}",
                ]
            )
        else:
            path_def = ""

        self.d = path_def


class Angled(Leaderline):
    """Leaderline comprised of one or two sharp 90 degree corners."""

    def route(self, origin, destination):

        start, end = self.end_points(origin, destination)

        if self.direction == "vh":
            # start vertical, end horizontal
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {start.x} {end.y}",
                    f"L {end.x} {end.y}",
                ]
            )
        elif self.direction == "hv":
            # start horizontal, end vertical
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {end.x} {start.y}",
                    f"L {end.x} {end.y}",
                ]
            )
        elif self.direction == "hh":
            # start horizontal, end horizontal
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {start.x + (end.x - start.x)/4} {start.y}",
                    f"L {start.x + (end.x - start.x)/4} {end.y}",
                    f"L {end.x} {end.y}",
                ]
            )
        elif self.direction == "vv":
            # start vertical, end vertical
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {start.x} {start.y + (end.y - start.y)/4}",
                    f"L {end.x} {start.y + (end.y - start.y)/4}",
                    f"L {end.x} {end.y}",
                ]
            )
        else:
            path_def = ""

        self.d = path_def


class Diagonal(Leaderline):
    """Leaderline comprised of a diagonal and horizontal line"""

    def route(self, origin, destination):

        start, end = self.end_points(origin, destination)
        x_dist = abs(start.x - end.x)
        y_dist = abs(start.y - end.y)

        min_dist = min(x_dist, y_dist)

        segment = min_dist / 6
        segment_x = segment * math.cos(math.radians(45))
        segment_y = segment * math.sin(math.radians(45))
        radius = segment * math.tan(math.radians(67.5))

        path_def = " ".join(
            [
                f"M {start.x} {start.y}",
                f"L {start.x + min_dist - segment_x} {start.y + min_dist - segment_y}",
                f"A {radius} {radius} 0 0 0  {start.x + min_dist + segment} {end.y}",
                f"L {end.x} {end.y}",
            ]
        )

        self.d = path_def


class DiagonalAngled(Leaderline):
    """Leaderline comprised of a diagonal and horizontal line"""

    def route(self, origin, destination):

        start, end = self.end_points(origin, destination)
        x_dist = abs(start.x - end.x)
        y_dist = abs(start.y - end.y)
        min_dist = min(x_dist, y_dist)

        path_def = " ".join(
            [f"M {start.x} {start.y}", f"l {min_dist} {min_dist}" f"L {end.x} {end.y}"]
        )

        self.d = path_def


class Straight(Leaderline):
    """Leaderline comprised of a single straight line."""

    def route(self, origin, destination):

        start, end = self.end_points(origin, destination)

        if self.direction == "vh":
            # start vertical, end horizontal
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {end.x} {end.y}",
                ]
            )
        elif self.direction == "hv":
            # start horizontal, end vertical
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {end.x} {end.y}",
                ]
            )
        elif self.direction == "hh":
            # start horizontal, end horizontal
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {end.x} {end.y}",
                ]
            )
        elif self.direction == "vv":
            # start vertical, end vertical
            path_def = " ".join(
                [
                    f"M {start.x} {start.y}",
                    f"L {end.x} {end.y}",
                ]
            )
        else:
            path_def = ""

        self.d = path_def



################################
#
# Default component settings
#
################################

# Pinlabel
pinlabel = {
    "tag": "pinlabel",
    "body": {
        "x": 6,
        "y": 0,
        "width": 80,
        "height": 26,
        "corner_radius": 3,
        "tag": "pinlabel__body",
    },
    "leaderline": {
        "direction": "hh",
        "tag": "pinlabel__leader",
    },
    "text": {
        "tag": "pinlabel__text",
    },
}

# Legend
legend = {
    "max_height": None,
    "inset": (10, 10, 10, 10),
    "tag": "legend",
    "entry": {
        "width": 159,
        "height": 28,
        "swatch": {
            "width": 20,
            "height": 20,
            "tag": "swatch",
        },
        "tag": "legendentry",
    },
}

# TextBlock
textblock = {
    "line_height": 22,
    "width": None,
    "height": None,
    "offset": (0, 0),
    "tag": "textblock",
}

# Annotation
annotation = {
    "tag": "annotation",
    "content": {
        "tag": "annotation__text",
        "x": 28,
        "y": 17,
        "line_height": 16,
    },
    "body": {
        "x": 40,
        "y": 29,
        "width": 250,
        "height": 50,
        "corner_radius": 25,
        "tag": "annotation__body",
    },
    "target": {
        "x": -10,
        "y": -10,
        "width": 20,
        "height": 20,
        "corner_radius": 10,
        "tag": "annotation__target",
    },
    "leaderline": {
        "direction": "vh",
        "tag": "annotation__leaderline",
    },
}

# Panel
panel = {
    "inset": (2, 2, 2, 2),
    "tag": "panel",
    "inner": {"tag": "panel__inner"},
    "outer": {"tag": "panel__outer"},
}


# Integrated circuit
ic_dip = {
    "inset": (15, 0, 15, 0),
    "tag": "ic ic--dip",
    "body": {
        "x": 15,
        "y": 0,
        "corner_radius": 3,
        "tag": "ic__body",
    },
    "leg": {
        "tag": "ic__leg",
    },
    "polarity_mark": {
        "radius": 5,
        "tag": "polarity",
    },
}
ic_qfp = {
    "inset": (15, 15, 15, 15),
    "pin_pitch": 30,
    "tag": "ic ic--qfp",
    "body": {
        "x": 15,
        "y": 15,
        "corner_radius": 3,
        "tag": "ic__body",
    },
    "leg": {
        "tag": "ic__leg",
    },
    "polarity_mark": {
        "radius": 5,
        "tag": "polarity",
    },
}

# Diagram layout template presets
diagram_presets = {
    "tag": "layout",
    "panel_00": {
        "inset": (2, 2, 2, 2),
        "tag": "panel",
        "inner": {"tag": "panel__inner"},
        "outer": {"tag": "panel__outer"},
    },
    "panel_01": {
        "inset": (0.5, 0.5, 0.5, 0.5),
        "tag": "panel--main",
        "inner": {"tag": "panel__inner"},
        "outer": {"tag": "panel__outer"},
    },
    "panel_02": {
        "inset": (0.5, 0.5, 0.5, 0.5),
        "tag": "panel--info",
        "inner": {"tag": "panel__inner"},
        "outer": {"tag": "panel__outer"},
    },
}


################################
#
# KiCad footprint settings
#
################################
kicad_6_footprints = {
    "version": 6,
    "layer": "User.1",
    "pinlabel": {
        "hide_fp_text_reference": True,
        "hide_fp_text_user": True,
        "value_offset": (0, 25),  # (mm dimensions)
    },
    "annotation": {
        "hide_fp_text_reference": True,
        "hide_fp_text_user": True,
        "value_offset": (25, 25),  # (mm dimensions)
    },
    "textblock": {
        "hide_fp_text_reference": True,
        "hide_fp_text_user": True,
    },
}
kicad_5_footprints = {
    "version": 5,
    "layer": "Eco1.User",
    "pinlabel": {
        "hide_fp_text_reference": True,
        "hide_fp_text_user": True,
        "value_offset": (0, 25),  # (mm dimensions)
    },
    "annotation": {
        "hide_fp_text_reference": True,
        "hide_fp_text_user": True,
        "value_offset": (25, 25),  # (mm dimensions)
    },
    "textblock": {
        "hide_fp_text_reference": True,
        "hide_fp_text_user": True,
    },
}



import copy
from pinout import core
from pinout.components.text import TextBlock
from pinout.components import leaderline as lline
from pinout.components.layout import Group

from pinout import config


class Leaderline(lline.Curved):
    pass


class Target(core.Rect):
    pass


class Body(core.Rect):
    pass


class Content(TextBlock):
    pass


class AnnotationLabel(Group):
    """Annotation style label."""

    def __init__(
        self,
        content=None,
        body=None,
        leaderline=None,
        target=None,
        **kwargs,
    ):
        self._content = None
        self._body = None
        self._leaderline = None
        self._target = None

        super().__init__(**kwargs)
        self.update_config(config.annotation)
        self.add_tag(self.config["tag"])

        self.leaderline = leaderline
        self.body = body
        self.target = target
        # content relied on body - must come after
        self.content = content

        # Route leaderline once other elements exist
        self._leaderline.route(self.target, self.body)

        self.add(self.leaderline)
        self.add(self.target)
        self.add(self.body)
        self.add(self.content)

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, content):
        content = copy.deepcopy(content or {})
        config = self.config["content"]
        # Parse content: str > list > dict > TextBlock
        if type(content) is str:
            content = [content]
        if type(content) is list:
            content = {
                "content": content,
                "x": self.body.x + self.body.width / 2,
                "y": (self.body.y + self.body.height / 2)
                - (config["line_height"] * (len(content) - 1) / 2 * self.scale.y),
            }
        if isinstance(content, dict):
            config.update(content)
            content = Content(**config)
        content.add_tag(self.config["content"]["tag"])
        content.scale = self.scale
        self._content = content

    @property
    def leaderline(self):
        return self._leaderline

    @leaderline.setter
    def leaderline(self, leaderline):
        leaderline = copy.deepcopy(leaderline or {})
        if isinstance(leaderline, dict):
            leaderline_config = self.config["leaderline"]
            leaderline_config.update(leaderline)
            leaderline = Leaderline(**leaderline_config)
        leaderline.add_tag(self.config["leaderline"]["tag"])
        self._leaderline = leaderline

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, target):
        target = copy.deepcopy(target or {})
        if isinstance(target, dict):
            target_config = self.config["target"]
            target_config.update(target)
            target = Target(**target_config)
        target.add_tag(self.config["target"]["tag"])
        self._target = target

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, body):
        body = copy.deepcopy(body or {})
        if isinstance(body, dict):
            body_config = self.config["body"]
            body_config.update(body)
            body = Body(**body_config)
        body.add_tag(self.config["body"]["tag"])
        self._body = body



from jinja2 import Environment, PackageLoader, select_autoescape


env = Environment(
    loader=PackageLoader("pinout", "templates"),
    autoescape=select_autoescape(["html", "xml"]),
    trim_blocks=True,
    lstrip_blocks=True,
)


def get(template_name):
    return env.get_template(template_name)



import copy
from pinout.core import SvgShape, Group, Rect, Text, BoundingCoords, Coords
from pinout.components import leaderline as lline
from pinout import config


def pitch_generator(start, pitch):
    x = start[0]
    y = start[1]
    while True:
        yield (x, y)
        x += pitch[0]
        y += pitch[1]


class Body(SvgShape):
    """Graphical shape that makes up the body of a pinlabel."""

    def __init__(self, x, y, width, height, corner_radius=0, **kwargs):
        self.corner_radius = corner_radius
        super().__init__(x=x, y=y, width=width, height=height, **kwargs)

    def bounding_coords(self):
        # PinLabelBody origin is vertically centered
        return BoundingCoords(
            self.x,
            self.y - (self.height / 2),
            self.x + self.width,
            self.y + (self.height / 2),
        )

    def render(self):
        body = Rect(
            x=self.x,
            y=self.y - (self.height / 2),
            width=self.width,
            height=self.height,
            corner_radius=self.corner_radius,
        )
        body.add_tag(config.pinlabel["body"]["tag"])
        return body.render()


class Leaderline(lline.Curved):
    """Graphical line joining the label origin coordinates to the label body."""

    pass


class Base(Group):
    """Label component designed specifically for labelling pins."""

    def __init__(
        self,
        content="",
        x=0,
        y=0,
        tag=None,
        body=None,
        leaderline=None,
        **kwargs,
    ):
        self.content = content
        self._leaderline = None
        self._body = None
        super().__init__(x, y, tag=tag, **kwargs)
        self.update_config(config.pinlabel)

        self.body = body
        self.leaderline = leaderline

        # Add leaderline and body reference into children
        self.add(self._body)

        # Add leaderline at render as it is replaced by pinlabelGroup!!!
        # Add SvgShape so pin label reports correct dimensions.
        self.add(SvgShape(x=self.leaderline.x, y=self.leaderline.x))

        self.add_tag(config.pinlabel["tag"])

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, body):
        # ensure instance data is unique
        body = copy.deepcopy(body or self.config["body"])
        # Convert dict into body object
        if isinstance(body, dict):
            body_config = self.config["body"]
            body_config.update(body)
            body = Body(**body_config)
            # Add body config tag if not there
        body.add_tag(self.config["body"]["tag"])
        self._body = body

    @property
    def leaderline(self):
        return self._leaderline

    @leaderline.setter
    def leaderline(self, leaderline):
        # ensure instance data is unique
        leaderline = copy.deepcopy(leaderline or self.config["leaderline"])
        # Convert dict into leaderline object
        if isinstance(leaderline, dict):
            leaderline_config = self.config["leaderline"]
            leaderline_config.update(leaderline)
            leaderline = Leaderline(**leaderline_config)
        # Add leaderline config tag if not there
        leaderline.add_tag(self.config["leaderline"]["tag"])
        self._leaderline = leaderline

    def render(self):
        # Add text content
        x = self.body.width / 2 + self.body.x
        y = self.body.y
        self.add(
            Text(
                self.content,
                x=x,
                y=y,
                tag=config.pinlabel["text"]["tag"],
                scale=self.scale,
            )
        )
        # Route leaderline
        self.leaderline.route(Rect(), self._body)
        self.add(self.leaderline)
        return super().render()


class PinLabel(Base):
    pass


class PinLabelGroup(Group):
    """Convenience class to place multiple rows of pin-labels on a pin-header."""

    def __init__(
        self,
        x,
        y,
        pin_pitch,
        label_start,
        label_pitch,
        labels,
        leaderline=None,
        body=None,
        **kwargs,
    ):
        scale = Coords(*kwargs.pop("scale", (1, 1)))
        super().__init__(x=x, y=y, **kwargs)

        # Setup generators for row locations
        pin_coords = pitch_generator((0, 0), pin_pitch)
        label_coords = pitch_generator(label_start, label_pitch)

        for row in labels:
            row_group = self.add(Group())
            for label in row:

                # If data is supplied convert to Label
                if type(label) is tuple:
                    content, tag, *args = label
                    attrs = args[0] if len(args) > 0 else {}

                    # Set leaderline and body in attrs if supplied in either:
                    # 1. data
                    # 2. PinlabelGroup
                    attrs["leaderline"] = attrs.get("leaderline", None) or leaderline
                    attrs["body"] = attrs.get("body", None) or body

                    label = PinLabel(
                        content=content,
                        scale=scale,
                        **attrs,
                    )

                # -- label now exists -- #
                label.add_tag(tag)

                # Label follows another label in the row
                try:
                    prev_label = row_group.children[-1]
                    label.x = prev_label.x + prev_label.width * scale.x
                    label.y = prev_label.y + prev_label.body.y * scale.y
                    label.leaderline = lline.Straight(direction="hh")
                # Start of a new row
                except IndexError:
                    label.x, label.y = next(pin_coords)
                    x, y = next(label_coords)

                    label.body.x += x - label.x * scale.x
                    label.body.y += y - label.y * scale.y

                row_group.add(label)



import base64
import copy
import math
import pathlib
import PIL
import re
import urllib.request
import uuid
import warnings
import xml.etree.ElementTree as ET
from collections import namedtuple
from pinout import manager, templates


Coords = namedtuple("Coords", ("x y"))
BoundingCoords = namedtuple("BoundingCoords", ("x1 y1 x2 y2"))
BoundingRect = namedtuple("BoundingCoords", ("x y w h"))


class IdGenerator:
    def __init__(self):
        self.base = str(uuid.uuid4())
        self.counter = 0

    def __call__(self):
        self.counter += 1
        return f"{self.base}_{self.counter}"


diagram_id = IdGenerator()


class TransformMixin:
    def __init__(
        self,
        matrix=None,
        translate=None,
        scale=(1, 1),
        rotate=0,
        skewx=None,
        skewy=None,
        **kwargs,
    ):
        # kwargs that make it to here are ignored.
        self.matrix = matrix
        self.translate = translate
        self.scale = Coords(*scale)
        self.rotate = rotate
        self.skewx = skewx
        self.skewy = skewy

        super().__init__()


class Component:
    """common functions and attributes shared by all components"""

    def __init__(self, clip=None, config=None, defs=None, tag=None, **kwargs):
        self._clip = None
        self.config = config or {}
        self.defs = defs or []
        self.id = diagram_id()
        self.tag = tag
        super().__init__(**kwargs)

        self.clip = clip

    @property
    def clip(self):
        return self._clip

    @clip.setter
    def clip(self, obj):
        if obj:
            if isinstance(obj, ClipPath):
                self._clip = obj
            else:
                self._clip = ClipPath(children=obj)

    def add_def(self, instance):
        """Add a component to the svg 'def' section"""
        if instance:
            self.defs.append(instance)
        return instance

    def add_tag(self, tag):
        """Append a tag to the instance

        :param tag: CSS class name(s)
        :type tag: string
        """
        tag_list = (self.tag or "").split(" ")
        if tag not in tag_list:
            tag_list.append(tag)
        self.tag = " ".join(tag_list)

    def update_config(self, vals):
        """update config dict

        :param vals: Values to update
        :type vals: dict
        """
        self.config.update(copy.deepcopy(vals))

    def render_defs(self):
        """Render SVG markup from 'defs'

        :return: SVG markup
        :rtype: string
        """
        content = ""

        # Transfer clippath to defs
        self.add_def(self.clip)

        # Render defs
        for d in self.defs:
            try:
                content += d.render()
            except AttributeError as e:
                print("*", d)
                print(d.render)
                print(d.render())
                print(e)

        # Recursively render children defs
        try:
            for child in [
                child for child in self.children if hasattr(child, "render_defs")
            ]:
                content += child.render_defs()
        except AttributeError:
            pass
            # Object has no children

        return content

    @staticmethod
    def rotate_box_coords(origin, coords, rotate):
        ox, oy = origin
        x1, y1, x2, y2 = coords
        # rotate corners of bounding box
        corners = [(x1, y1), (x2, y1), (x1, y2), (x2, y2)]
        rx = []
        ry = []
        for (x, y) in corners:
            rx.append(
                ox
                + (x - ox) * math.cos(math.radians(rotate))
                - (y - oy) * math.sin(math.radians(rotate))
            )
            ry.append(
                oy
                + (x - ox) * math.sin(math.radians(rotate))
                + (y - oy) * math.cos(math.radians(rotate))
            )

        return BoundingCoords(min(rx), min(ry), max(rx), max(ry))


class Layout(Component, TransformMixin):
    """Base class fundamentally grouping other components together."""

    def __init__(self, x=0, y=0, children=None, **kwargs):
        self.x = x
        self.y = y
        self.children = children or []

        super().__init__(**kwargs)

    def add(self, instance):
        self.children.append(instance)
        return instance

    @staticmethod
    def find_children_by_type(component, target_type):
        """Recursively find all children of the component and it's decendents by type.

        :param component: Component instance to start seach from
        :type component: class instance
        :param target_type: class  to match with instances
        :type target_type: class
        :return: All instances of type 'target_type' that are descendents of 'component'
        :rtype: list
        """
        results = []
        try:
            for c in component.children:
                if isinstance(c, target_type):
                    results.append(c)
                results += Layout.find_children_by_type(c, target_type)
        except AttributeError:
            pass
            # No children

        return results

    def bounding_rect(self):
        """Top left coordinates with width and height of a bounding rectangle

        :return: origin coordinates and rectangle dimensions
        :rtype: tuple (x, y, w, h)
        """
        x1, y1, x2, y2 = self.bounding_coords()
        return BoundingRect(x1, y1, x2 - x1, y2 - y1)

    def bounding_coords(self):
        """Coordinates of the components's bounding rectangle.

        :return: (x1, y1, x2, y2)
        :rtype: BoundingCoords (namedtuple)
        """
        # Collect bounding coords of children
        x = []
        y = []
        if self.clip:
            targets = self.clip.children
        else:
            targets = self.children
        for child in [
            instance
            for instance in targets
            if hasattr(type(instance), "bounding_coords")
        ]:
            coords = child.bounding_coords()
            x.append(self.x + coords.x1 * self.scale.x)
            y.append(self.y + coords.y1 * self.scale.y)
            x.append(self.x + coords.x2 * self.scale.x)
            y.append(self.y + coords.y2 * self.scale.y)

        try:
            x1, y1, x2, y2 = min(x), min(y), max(x), max(y)

            # Clip object has its own rotation
            if self.clip:
                return BoundingCoords(x1, y1, x2, y2)

            # rotate corners of bounding box
            corners = [(x1, y1), (x2, y1), (x1, y2), (x2, y2)]
            rx = []
            ry = []
            for (x, y) in corners:
                rx.append(
                    self.x
                    + (x - self.x) * math.cos(math.radians(self.rotate))
                    - (y - self.y) * math.sin(math.radians(self.rotate))
                )
                ry.append(
                    self.y
                    + (x - self.x) * math.sin(math.radians(self.rotate))
                    + (y - self.y) * math.cos(math.radians(self.rotate))
                )

            return BoundingCoords(min(rx), min(ry), max(rx), max(ry))

        except ValueError:
            # There are no children
            return BoundingCoords(0, 0, 0, 0)

    def render_children(self):
        """Render SVG markup from 'children'

        :return: SVG markup
        :rtype: string
        """
        content = ""
        for child in self.children:
            try:
                content += child.render()
            except TypeError as e:
                print(f"An error occurred rendering: {child}")
                print(child.render())
                print(e)
        return content


class Use(Layout):
    """Implement <use> svg tag"""

    def __init__(self, instance, **kwargs):
        self.target_id = instance.id

        super().__init__(**kwargs)

    def render(self):
        # convert kwargs into parameters for <use>
        tplt = templates.get("use.svg")
        return tplt.render(use=self)


class Group(Layout):
    """Group components together"""

    def __init__(self, x=0, y=0, tag=None, **kwargs):
        super().__init__(x=x, y=y, tag=tag, **kwargs)

    @property
    def width(self):
        return self.bounding_rect().w

    @property
    def height(self):
        return self.bounding_rect().h

    def render(self):
        """Render children into a <group> tag.

        :return: SVG markup
        :rtype: string
        """
        tplt = templates.get("group.svg")
        return tplt.render(group=self)


class ClipPath(Group):
    """Define a clip-path component"""

    def __init__(self, children=None, **kwargs):
        super().__init__(**kwargs)
        # Accept 'children' as list or single instance.
        children = children or []
        try:
            for child in children:
                self.add(child)
        except TypeError:
            # Children is a single component (not an iterible)
            self.add(children)

    def render(self):
        """Render children into a <clipPath> tag."""
        tplt = templates.get("clippath.svg")
        return tplt.render(path=self)


class StyleSheet:
    """Include a cascading stylesheet."""

    def __init__(self, src, embed=False):
        self._src = None
        self.src = src
        self.embed = embed

    @property
    def src(self):
        return self._src

    @src.setter
    def src(self, val):
        self._src = pathlib.Path(val)

    def render(self):
        tplt = templates.get("style.svg")
        if not self.embed:
            return tplt.render(stylesheet=self)
        else:
            data = manager.load_data(self.src)
            return tplt.render(data=data)


class Raw:
    """Include arbitary code to the document"""

    def __init__(self, content):
        self.content = content

    def render(self):
        return self.content


class SvgShape(Component, TransformMixin):
    """Base class for components that have a graphical representation."""

    def __init__(self, x=0, y=0, width=0, height=0, **kwargs):
        self.x = x
        self.y = y
        self._width = width
        self._height = height

        super().__init__(**kwargs)

    @property
    def width(self):
        if self.clip:
            return self.clip.width
        return self._width

    @width.setter
    def width(self, val):
        self._width = val

    @property
    def height(self):
        if self.clip:
            return self.clip.height
        return self._height

    @height.setter
    def height(self, val):
        self._height = val

    def bounding_rect(self):
        """Component's origin coordinates and dimensions"""
        x1, y1, x2, y2 = self.bounding_coords()
        return BoundingRect(x1, y1, x2 - x1, y2 - y1)

    def bounding_coords(self):
        """Coordinates representing a shape's bounding-box."""
        if self.clip:
            return self.clip.bounding_coords()
        else:
            x = [self.x * self.scale.x, (self.x + self.width) * self.scale.x]
            y = [self.y * self.scale.y, (self.y + self.height) * self.scale.y]

            return Component.rotate_box_coords(
                origin=(self.x, self.y),
                coords=(min(x), min(y), max(x), max(y)),
                rotate=self.rotate,
            )

    def render(self):
        return ""


class Path(SvgShape):
    """SVG Path object"""

    def __init__(self, path_definition="", **kwargs):
        super().__init__(**kwargs)
        self.d = path_definition

    def render(self):
        """Render a <path> tag.

        :return: SVG markup
        :rtype: string
        """
        tplt = templates.get("path.svg")
        return tplt.render(path=self)


class Rect(SvgShape):
    """SVG <rect> object"""

    def __init__(self, *args, corner_radius=0, **kwargs):
        self.corner_radius = corner_radius
        super().__init__(*args, **kwargs)

    def render(self):
        """Render a <rect> tag.

        :return: SVG markup
        :rtype: string
        """
        tplt = templates.get("rect.svg")
        return tplt.render(rect=self)


class Circle(SvgShape):
    """SVG <circle> object"""

    def __init__(self, cx, cy, r, **kwargs):
        self.r = r
        kwargs["x"] = cx
        kwargs["y"] = cy
        super().__init__(**kwargs)

    def render(self):
        """Render a <circle> tag."""

        tplt = templates.get("circle.svg")
        return tplt.render(circle=self)


class Text(SvgShape):
    """SVG <text> object"""

    def __init__(self, content, **kwargs):
        super().__init__(**kwargs)
        self.content = content

    def render(self):
        """Render a <text> tag.

        :return: SVG markup
        :rtype: string
        """
        tplt = templates.get("text.svg")
        return tplt.render(text=self)


class Image(SvgShape):
    """Include an image in the diagram."""

    def __init__(self, src, dpi=72, embed=False, **kwargs):
        self.coords = kwargs.pop("coords", {})
        self._dpi = dpi
        self.embed = embed
        self.im_size = (1, 1)
        self.src = src
        self.svg_data = None

        try:
            # Load image dimensions to avoid multiple loads when calculating coords
            # Allow relative paths outside CWD
            cwd = pathlib.Path.cwd()
            im = PIL.Image.open(cwd.joinpath(self.src))
            self.im_size = im.size
        except TypeError as e:
            # Image src is assumed to be another Image instance
            self.im_size = self.src.im_size
            self.coords = self.src.coords

        except PIL.UnidentifiedImageError:
            # Image is assumed to be SVG
            self.set_svg_im_size()

        except OSError:
            try:
                # file not at local path, try path as URL
                im = PIL.Image.open(urllib.request.urlopen(self.src))
                self.im_size = im.size
            except (PIL.UnidentifiedImageError):
                # Image is assumed to be SVG
                # Match arbitary im_size to width and height so no scaling occurs
                pass

        kwargs["width"] = kwargs.get("width", self.im_size[0])
        kwargs["height"] = kwargs.get("height", self.im_size[1])
        super().__init__(**kwargs)

    @property
    def dpi(self):
        return self._dpi

    @dpi.setter
    def dpi(self, val):
        self._dpi = val
        self.set_svg_im_size()

    @property
    def src(self):
        return self._src

    @src.setter
    def src(self, val):
        # Ensure src is either a pathlib.Path or Image instance
        if isinstance(val, Image):
            self._src = val
        else:
            self._src = pathlib.Path(val)

    def set_svg_im_size(self):
        # Extract dimensions from SVG attributes
        tree = ET.parse(self.src)
        root = tree.getroot()
        try:
            width = root.attrib["width"]
            height = root.attrib["height"]
        except KeyError:
            # SVG can omit width and height.
            # Use viewBox dimensions instead.
            width, height = root.attrib["viewBox"].split(" ")[-2:]
        # Dimensions may (or may not) include units
        # re splits at start and end of matched group hence x3 vars
        r = re.compile(r"(^[\d\.]+)")
        _, width, width_units = re.split(r, width)
        _, height, height_units = re.split(r, height)
        # Convert to pixel dimensions
        width = self.to_pixels(width, width_units)
        height = self.to_pixels(height, height_units)
        # Set im_size
        self.im_size = (width, height)

    def to_pixels(self, value, units):

        value = float(value)

        if not units or units.strip().lower() == "px":
            return value

        elif units.strip().lower() == "mm":
            return value / 25.4 * self.dpi

        elif units.strip().lower() == "cm":
            return value / 2.54 * self.dpi

        elif units.strip().lower() == "in":
            return value * self.dpi

        elif units.strip().lower() == "pt":
            return value * 72 * self.dpi

        else:
            warnings.warn(
                f'Image dimension "{units}" is not recognised. Leaving unchanged.'
            )
            return value

    def coord(self, name, raw=False):
        """Return scaled coordinatates."""

        x, y = self.coords[name]

        # Actual image size:
        iw, ih = self.im_size

        # Scale x and y to match user supplied dimensions
        # NOTE: use *_width* and *_height* to ensure actual width and not clipped width
        scaler = min(self._width / iw, self._height / ih)

        # Transformed size
        tw, th = iw * scaler, ih * scaler

        # Transform x and y coords
        tx = x * scaler
        ty = y * scaler

        if not raw:
            # if 'raw' is False translate the coords
            # NOTE: SVG transforms images proportionally to 'fit' supplied dimensions

            # Calculate offset to centre fitted image
            tx = tx + (self._width - tw) / 2
            ty = ty + (self._height - th) / 2

            # rotate coords
            rtx = tx * math.cos(math.radians(self.rotate)) - ty * math.sin(
                math.radians(self.rotate)
            )
            rty = tx * math.sin(math.radians(self.rotate)) + ty * math.cos(
                math.radians(self.rotate)
            )
            tx = rtx + self.x
            ty = rty + self.y

        return Coords(tx, ty)

    def add_coord(self, name, x, y):
        """Record coordinates of the **unscaled** image."""
        self.coords[name] = Coords(x, y)

    def loadData(self):
        """Load image data from URL or local file system."""
        try:
            with open(self.src, "rb") as f:
                return f.read()
        except OSError:
            try:
                with urllib.request.urlopen(self.src) as f:
                    return f.read()
            except urllib.error.HTTPError as e:
                print(e.code)

    def render(self):
        """Render SVG markup either linking or embedding an image."""
        if isinstance(self.src, Image):
            # src image is wrapped in <use> tag which has to replicate 'fitting' behaviour of SVG images
            scaler = min(self._width / self.src._width, self._height / self.src._height)
            self.scale = Coords(
                self.scale.x * scaler,
                self.scale.y * scaler,
            )
            actual_width = self.src._width * scaler
            actual_height = self.src._height * scaler

            tx = (self._width - actual_width) / 2
            ty = (self._height - actual_height) / 2

            # Rotate coords
            rtx = tx * math.cos(math.radians(self.rotate)) - ty * math.sin(
                math.radians(self.rotate)
            )
            rty = tx * math.sin(math.radians(self.rotate)) + ty * math.cos(
                math.radians(self.rotate)
            )

            # clip-path must be a separate component when using <use> to
            # avoid applying scale to clip-path.
            output = Group(clip=self.clip)
            # Reference image from defs with <use> tag
            output.add(
                Use(
                    x=self.x + rtx,
                    y=self.y + rty,
                    scale=self.scale,
                    tag=self.tag,
                    instance=self.src,
                    rotate=self.rotate,
                )
            )

            return output.render()

        # Use externally referenced image
        media_type = pathlib.Path(self.src).suffix[1:]
        tplt = templates.get("image.svg")

        if self.embed:
            if media_type == "svg":
                data = self.loadData()
                # Extract JUST the <svg> markup with no <XML> tag
                tree = ET.fromstring(data)
                self.svg_data = ET.tostring(tree)
            else:
                encoded_img = base64.b64encode(self.loadData())
                # IMPORTANT: bypass Image.src setter function
                self._src = (
                    f"data:image/{media_type};base64,{encoded_img.decode('utf-8')}"
                )
        return tplt.render(image=self)



from collections import namedtuple
import copy
import math
import tokenize
from tokenize import ENDMARKER, LPAR, RPAR, NAME, NUMBER, OP
import re
import warnings

from pinout.components.pinlabel import PinLabelGroup
from pinout.components.annotation import AnnotationLabel

Pinout_item = namedtuple("Pinout_item", ["content", "attrs", "x", "y", "scale"])


def format_tag_warning(message, category, filename, lineno, file=None, line=None):
    return f"{category.__name__}: {message}"


def rotate_point_coord(coord, rotate):
    """Rotate a coordinate around (0,0) by angle in degrees."""
    # Rotate in opposite direction to kicad for SVG coordinate space
    rotate = -rotate
    x, y = coord
    x1 = x * math.cos(math.radians(rotate)) - y * math.sin(math.radians(rotate))
    y1 = x * math.sin(math.radians(rotate)) + y * math.cos(math.radians(rotate))
    return (x1, y1)


class Counter:
    def __init__(self, start_num):
        self.count = start_num

    def __call__(self):
        self.count += 1
        return self.count


class Tokens:
    def __init__(self, filepath):
        self.tokens = self.load_tokens(filepath)
        self.depth = 0
        self.current = None
        # Get initial value
        self.next()

    def next(self):
        self.current = next(self.tokens)
        self.update_depth(self.current)
        return self.current

    def goto(self, name):
        while self.current.string != name:
            self.next()
        return True

    def load_tokens(self, filepath):
        with tokenize.open(filepath) as f:
            tokens = tokenize.generate_tokens(f.readline)
            for tok in tokens:
                yield tok

    def update_depth(self, token):
        if token.exact_type == LPAR:
            self.depth += 1
        elif token.exact_type == RPAR:
            self.depth -= 1


class KiCadParser:
    def __init__(self, pcb_file, dpi=72, version=6):
        self.dpi = dpi
        self.filepath = pcb_file
        self.version = version

    def fp_text(self, tokens):
        exit_depth = tokens.depth
        fp_text = {
            "type": tokens.next().string.strip('"'),
            "text": tokens.next().string.strip('"'),
        }
        while tokens.depth >= exit_depth:
            if tokens.current.type == NAME:
                if tokens.current.string == "at":
                    fp_text["at"] = self.at(tokens)
                if tokens.current.string == "layer":
                    fp_text["layer"] = tokens.next().string.strip('"')
            tokens.next()
        return fp_text

    def fp_text_kicad_5(self, tokens):
        exit_depth = tokens.depth
        fp_text = {
            "type": tokens.next().string,
            "text": "",
            "layer": "",
        }
        # collate text until attribute starts
        while tokens.next().exact_type != LPAR:
            fp_text["text"] += tokens.current.string.strip('"')

        # Find and collate attributes
        tokens.next()
        while tokens.depth >= exit_depth:
            if tokens.current.type == NAME:
                if tokens.current.string == "at":
                    fp_text["at"] = self.at(tokens)
                if tokens.current.string == "layer":
                    while tokens.current.exact_type != RPAR:
                        fp_text["layer"] += tokens.next().string
            tokens.next()
        return fp_text

    def at(self, tokens):

        # 'at' can be 2 to 3 numbers. x, y, (and optional) rotation
        at = []
        prefix = ""
        exit_depth = tokens.depth
        while tokens.depth >= exit_depth:
            if tokens.current.type == NUMBER:
                at.append(prefix + tokens.current.string.strip('"'))
                prefix = ""
            elif tokens.current.type == OP:
                prefix = tokens.current.string
            tokens.next()

        # Convert strings into numbers
        at = [float(num) for num in at]

        # Convert x and y (mm) values to px.
        # !!! Do NOT convert rotation value
        at[:2] = [self.mm_to_dpi(val) for val in at[:2]]

        return at

    def mm_to_dpi(self, mm):
        return mm / 25.5 * self.dpi

    def get_single_attr(self, key):
        tokens = Tokens(self.filepath)
        tokens.goto(key)
        return tokens.next().string.strip('"')

    def version(self):
        return self.get_single_attr("version")

    def generator(self):
        return self.get_single_attr("generator")

    def paper(self):
        return self.get_single_attr("paper")

    def general(self):
        general = {}
        tokens = Tokens(self.filepath)
        tokens.goto("general")
        exit_depth = tokens.depth
        while tokens.depth >= exit_depth:
            # Assumes all entries are (key value) pairs
            if tokens.current.type == NAME and tokens.depth == exit_depth + 1:
                key = tokens.current.string.strip('"')
                val = tokens.next()
                if val.type == NUMBER:
                    general[key] = float(val.string)
                else:
                    general[key] = val.string.strip('"')
            tokens.next()
        return general

    def layers(self):
        layers = {}
        tokens = Tokens(self.filepath)
        tokens.goto("layers")
        exit_depth = tokens.depth
        while tokens.depth >= exit_depth:
            if tokens.current.type == NUMBER:
                key = int(tokens.current.string)
                layers[key] = {
                    "canonical_name": tokens.next().string.strip('"'),
                    "type": tokens.next().string.strip('"'),
                    "user_name": None,
                }
                if tokens.next().type != OP:
                    layers[key]["user_name"] = tokens.current.string.strip('"')
            tokens.next()
        return layers

    def footprints(self):
        # counter used give unique id to each footprint
        counter = Counter(0)
        footprints = []
        tokens = Tokens(self.filepath)

        # Set footprint name depending on KiCad version
        if self.version < 6:
            fooprint_name = "module"
        else:
            fooprint_name = "footprint"

        while tokens.current.type != ENDMARKER:
            if tokens.current.type == NAME and tokens.current.string == fooprint_name:
                fp = {
                    "id": counter(),
                    "at": None,
                    "fp_text": [],
                    "layer": "",
                    "name": tokens.next().string.strip('"'),
                }
                exit_depth = tokens.depth

                if self.version >= 6:

                    while tokens.depth >= exit_depth:
                        tokens.next()
                        if tokens.current.type == NAME:
                            if tokens.current.string == "layer":
                                fp["layer"] = tokens.next().string.strip('"')
                            if tokens.current.string == "at":
                                fp["at"] = self.at(tokens)
                            if tokens.current.string == "fp_text":
                                fp["fp_text"].append(self.fp_text(tokens))
                    footprints.append(fp)

                else:
                    #####################################################
                    # KiCad 5 does NOT wrap attribute strings in quotes #
                    # tokens need to be collated until end of attribute.#
                    #####################################################
                    while tokens.next().exact_type != LPAR:
                        fp["name"] += tokens.current.string

                    while tokens.depth >= exit_depth:
                        if (
                            tokens.next().type == NAME
                            and tokens.depth == exit_depth + 1
                        ):
                            if tokens.current.string == "layer":
                                while tokens.next().exact_type != RPAR:
                                    fp["layer"] += tokens.current.string
                            elif tokens.current.string == "at":
                                fp["at"] = self.at(tokens)
                            elif tokens.current.string == "fp_text":
                                fp["fp_text"].append(self.fp_text_kicad_5(tokens))

                            tokens.next()
                    footprints.append(fp)

            tokens.next()
        return footprints

    def gr_text(self):
        text_list = []
        tokens = Tokens(self.filepath)
        while tokens.next().type != ENDMARKER:
            if tokens.current.type == NAME and tokens.current.string == "gr_text":
                content = tokens.next().string.strip('"')
                text_list.append(content)
        return text_list


class PinoutParser(KiCadParser):
    """Filter and format data exclusively for pinout requirements"""

    def __init__(self, filepath, dpi=72, version=6):
        super().__init__(filepath, dpi, version)
        self.pinout_img = None

    def layers(self):
        layers = super().layers()
        pinout_layers = [
            layer["canonical_name"]
            for layer in layers.values()
            if layer["user_name"] and layer["user_name"].lower().startswith("pinout")
        ]
        return pinout_layers

    def link_image(self, image_instance):
        self.pinout_img = image_instance

    def footprints(self):
        pinout_fps = []
        pinout_origin = [0, 0]
        fps = super().footprints()
        # Collect all pinout footprints
        for fp in fps:
            if fp["name"].startswith("pinout"):
                pinout_fps.append(fp)

        # Find pinout origin
        # Needs to be a copy as original pinout_origin
        # is later changed to (0,0)
        for fp in pinout_fps:
            if fp["name"] == "pinout:Origin":
                pinout_origin = copy.copy(fp["at"])

        # Offset coordinates by pinout_origin
        for fp in pinout_fps:
            fp["at"][:2] = [a - b for a, b in zip(fp["at"][:2], pinout_origin)]

        # Rotate fp_text coords

        for fp in pinout_fps:
            try:
                rotate = fp["at"][2]
            except IndexError:
                rotate = 0

            for fp_text in fp["fp_text"]:
                x, y = fp_text["at"][:2]
                rx, ry = rotate_point_coord((x, y), rotate)
                fp_text["at"][:2] = [rx, ry]

        return pinout_fps

    def transfer_footprint_coords(self, footprint):

        """Transfer footprint and child fp_text coords to pinout Image instance"""

        self.pinout_img.add_coord(f"footprint_{footprint['id']}", *footprint["at"][:2])

        for fp_text in footprint["fp_text"]:
            if fp_text["type"] == "value":
                self.pinout_img.add_coord(
                    f"{fp_text['text']}_{footprint['id']}", *fp_text["at"][:2]
                )

    def extract_tags(self, content):
        tag = None
        match = re.search("{{[\.\-_ a-zA-Z0-9]+}}", content)
        if match:
            tag = match.group(0)[2:-2]
            content = re.sub("{{[\.\- _a-zA-Z0-9]+}}", "", content).strip('"').strip()
        return tag, content

    def get_scale(self, x, y):
        # Deduce scale from label location (it is relative to pin)
        try:
            sx = x / abs(x)
        except ZeroDivisionError:
            sx = 1
        try:
            sy = y / abs(y)
        except ZeroDivisionError:
            sy = 1
        return (sx, sy)

    def get_footprint_specs(self, footprint):
        for fp_text in footprint["fp_text"]:
            attrs = ["hh"]
            if fp_text["type"] == "user":
                # Excepted format: <leaderline direction>
                attrs = fp_text["text"].split(" ")
            elif fp_text["type"] == "value":
                x, y = self.pinout_img.coord(
                    f"{fp_text['text']}_{footprint['id']}", raw=True
                )
                scale = self.get_scale(x, y)
                content = fp_text["text"]
        return Pinout_item(content, attrs, x, y, scale)

    def gr_text(self):
        text = {}
        for content in super().gr_text():
            key, content = self.extract_tags(content)
            if key:
                # sub double quotes back in
                content = re.sub("{dblquote}", '"', content)
                # unescape chars
                content = bytes(content, "utf-8").decode("unicode_escape")
                text[key] = content
        return text

    def add_pinlabels(self, container):
        for fp in self.footprints():

            self.transfer_footprint_coords(fp)

            # Get transformed footprint coords
            pin_x, pin_y = self.pinout_img.coord(f"footprint_{fp['id']}")

            # Add pinlabels
            if fp["name"] == "pinout:PinLabel":
                self.transfer_footprint_coords(fp)

                # Get transformed footprint coords
                pin_x, pin_y = self.pinout_img.coord(f"footprint_{fp['id']}")

                fp_specs = self.get_footprint_specs(fp)

                # Parse label content
                tags = [
                    t[2:-2]
                    for t in re.findall(
                        "{{-?[_a-zA-Z]+[_a-zA-Z0-9-]*}}", fp_specs.content
                    )
                ]
                txts = [
                    t.strip()
                    for t in re.split(
                        "{{-?[_a-zA-Z]+[_a-zA-Z0-9-]*}}", fp_specs.content
                    )
                    if t
                ]
                if len(txts) > len(tags):
                    # User has fogotten to include tags on pinlabels
                    # Warn them and add place-holder tags
                    len_diff = len(txts) - len(tags)
                    txts += ["no_tag_assigned"] * len_diff

                    warnings.formatwarning = format_tag_warning
                    warnings.warn(
                        "pin-label(s) have not been assigned a tag in KiCad.",
                        SyntaxWarning,
                    )

                labels = list(zip(txts, tags))
                # Create pinlabel group
                container.add(
                    PinLabelGroup(
                        x=pin_x,
                        y=pin_y,
                        pin_pitch=(0, 0),
                        label_start=(abs(fp_specs.x), abs(fp_specs.y)),
                        label_pitch=(0, 0),
                        labels=[labels],
                        scale=fp_specs.scale,
                        leaderline={"direction": fp_specs.attrs[0]},
                    )
                )

    def add_annotations(self, container):
        for fp in self.footprints():

            if fp["name"] == "pinout:Annotation":
                scale = (1, 1)
                self.transfer_footprint_coords(fp)

                # Get transformed footprint coords
                pin_x, pin_y = self.pinout_img.coord(f"footprint_{fp['id']}")

                for fp_text in fp["fp_text"]:
                    # Get annotation attrs
                    if fp_text["type"] == "user":
                        # Excepted format: <leaderline direction>
                        direction, *_ = fp_text["text"].split(" ")

                    elif fp_text["type"] == "value":
                        x, y = self.pinout_img.coord(
                            f"{fp_text['text']}_{fp['id']}", raw=True
                        )
                        tag, content = self.extract_tags(fp_text["text"])

                        scale = self.get_scale(x, y)

                container.add(
                    AnnotationLabel(
                        tag=tag,
                        content=content,
                        x=pin_x,
                        y=pin_y,
                        body={"x": abs(x), "y": abs(y)},
                        scale=scale,
                        leaderline={"direction": direction},
                    )
                )



import uuid
from pinout import templates, config
from pinout.core import (
    Layout,
    StyleSheet,
    Group,
    SvgShape,
    Rect,
    BoundingCoords,
)


class Diagram(Layout):
    """Basis of a pinout diagram"""

    def __init__(self, width, height, tag=None, **kwargs):
        self.width = width
        self.height = height
        super().__init__(tag=tag, **kwargs)
        self.add(SvgShape(width=width, height=height))

    def add_stylesheet(self, path, embed=False):
        """Add a stylesheet to the diagram"""
        self.children.insert(0, StyleSheet(path, embed))

    def render(self):
        """Render children into an <svg> tag."""

        # Warn user if no styles have been added
        stylesheets = self.find_children_by_type(self, StyleSheet)
        if not stylesheets:
            print(
                """
        *********************
        No stylesheet is attached, the diagram may not appear as expected! 
        Generate one automatically with:
        >>> py -m pinout.manager --css <your_script_name> styles.css

        More info at:
        https://pinout.readthedocs.io/en/latest/pages/manager.html#generate-a-cascading-stylesheet
        *********************
                """
            )

        tplt = templates.get("svg.svg")
        return tplt.render(svg=self)


class Panel(Layout):
    def __init__(self, width, height, inset=None, **kwargs):
        """Assist with content grouping and positioning"""

        self.width = width
        self.height = height

        kwargs["config"] = kwargs.get("config", config.panel)

        super().__init__(**kwargs)

        inset = inset or self.config["inset"]
        self.inset = BoundingCoords(*inset)
        self.add_tag(config.panel["tag"])

        # add a non-rendering shape so component
        # reports user set coordinates and dimensions
        self.add(
            SvgShape(
                x=-self.inset.x1,
                y=-self.inset.y1,
                width=width,
                height=height,
            )
        )

        # Offset component to align children with inner dimensions
        self.x += self.inset.x1
        self.y += self.inset.y1

    @property
    def inset_width(self):
        return self.width - (self.inset.x1 + self.inset.x2)

    @property
    def inset_height(self):
        return self.height - (self.inset.y1 + self.inset.y2)

    def render(self):
        """Panel renders children into a <group> tag."""

        self.children.insert(
            0,
            Rect(
                width=self.width - (self.inset.x1 + self.inset.x2),
                height=self.height - (self.inset.y1 + self.inset.y2),
                tag=config.panel["inner"]["tag"],
            ),
        )
        # Insert a rect filling the outer component dimensions
        self.children.insert(
            0,
            Rect(
                x=-self.inset.x1,
                y=-self.inset.y1,
                width=self.width,
                height=self.height,
                tag=config.panel["outer"]["tag"],
            ),
        )

        tplt = templates.get("group.svg")
        return tplt.render(group=self)


class Diagram_2Columns(Diagram):
    def __init__(self, width, height, gutter, tag, **kwargs):
        self.gutter = gutter
        super().__init__(width, height, tag, **kwargs)

        # Add/override config
        self.config = config.diagram_presets
        self.update_config(kwargs.get("config", {}))

        self.panel_00 = self.add(
            Panel(
                x=0,
                y=0,
                width=width,
                height=height,
                tag=self.config["panel_00"]["tag"],
                config=self.config["panel_00"],
            )
        )
        self.panel_01 = self.panel_00.add(
            Panel(
                x=0,
                y=0,
                width=self.gutter,
                height=self.panel_00.inset_height,
                tag=self.config["panel_01"]["tag"],
                config=self.config["panel_01"],
            )
        )
        self.panel_02 = self.panel_00.add(
            Panel(
                x=self.gutter,
                y=0,
                width=self.panel_00.inset_width - self.gutter,
                height=self.panel_00.inset_height,
                tag=self.config["panel_02"]["tag"],
                config=self.config["panel_02"],
            )
        )


class Diagram_2Rows(Diagram):
    def __init__(self, width, height, gutter, tag, **kwargs):
        self.gutter = gutter
        super().__init__(width, height, tag, **kwargs)

        # Add/override config
        self.config = config.diagram_presets
        self.update_config(kwargs.get("config", {}))

        self.add_tag(self.config["tag"])

        self.panel_00 = self.add(
            Panel(
                x=0,
                y=0,
                width=width,
                height=height,
                tag=self.config["panel_00"]["tag"],
                config=self.config["panel_00"],
            )
        )
        self.panel_01 = self.panel_00.add(
            Panel(
                x=0,
                y=0,
                width=self.panel_00.inset_width,
                height=self.gutter,
                tag=self.config["panel_01"]["tag"],
                config=self.config["panel_01"],
            )
        )
        self.panel_02 = self.panel_00.add(
            Panel(
                x=0,
                y=self.gutter,
                width=self.panel_00.inset_width,
                height=self.panel_00.inset_height - self.gutter,
                tag=self.config["panel_02"]["tag"],
                config=self.config["panel_02"],
            )
        )



import random


def luminace(color_component):
    """Luminance of an individual Red, Green, or Blue, color component.

    :param color_component: Value between 0 and 255 (inclusive)
    :type color_component: int
    :return: Luminance value of the color component
    :rtype: float
    """
    i = color_component / 255

    if i <= 0.03928:
        return i / 12.92
    else:
        return ((i + 0.055) / 1.055) ** 2.4


def relative_luminance(rgb):
    """Normalised luminance value of an RGB color.

    :param rgb: Tuple (or List) representing RGB value.
    :type rgb: tuple
    :return: Value between 0 and 1.
    :rtype: float
    """
    return (
        0.2126 * luminace(rgb[0])
        + 0.7152 * luminace(rgb[1])
        + 0.0722 * luminace(rgb[2])
    )


def unique_contrasting_rgb(ref_color):
    """Generate a psudo-random color that, compared to 'ref_color', has a contrast ratio greater than 3. This contrast value is the minimum value to pass WCAG AA contrast recommendation for UI components.

    :param ref_color: Tuple (or List) representing RGB value.
    :type ref_color: tuple
    :return: Tuple representing an RGB color.
    :rtype: tuple
    """
    contrast = 0
    unique = False
    while contrast < 3 or not unique:
        rgb = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

        color_one_luminance = relative_luminance(ref_color)
        color_two_luminance = relative_luminance(rgb)

        if color_one_luminance > color_two_luminance:
            light = color_one_luminance
            dark = color_two_luminance
        else:
            dark = color_one_luminance
            light = color_two_luminance

        contrast = (light + 0.05) / (dark + 0.05)
        unique = is_distinct_rbg(rgb)
    palette.append(rgb)
    return rgb


def is_distinct_rbg(rgb_color, threshold=60):
    r, g, b = rgb_color
    for (pr, pg, pb) in palette:
        diff = abs(r - pr) + abs(g - pg) + abs(b - pb)
        if diff < threshold:
            return False
    return True


def assign_color(tag_list, ref_color=(255, 255, 255)):
    """Generate a stylesheet using metrics from a diagram. Various styles are tailored by making a *best-guess* based on diagram component dimensions or a *lucky-guess* filtered by preset criteria. The output should be considered a boot-strapping step to styling a diagram ...unless you feel lucky!

    :param diagram: The Diagram object requiring styling
    :type diagram: Diagram
    :return: content of a css stylesheet with all required styles to display a diagram.
    :rtype: str
    """

    # Assign random-ish color to each tag
    return [(tag, unique_contrasting_rgb(ref_color)) for tag in tag_list]


# Store created color for reference
palette = []


# Pinout Manager: management and utility functions

import argparse
import collections.abc
import importlib
import os
from pathlib import Path
import pkg_resources
import sys

from pinout import core

try:
    import cairosvg
except ImportError as e:
    pass
    # export as other (non svg) formats is not available

# user variables
# C:\Program Files\mingw-w64\x86_64-8.1.0-posix-seh-rt_v6-rev0\mingw64\bin

# env variables
# C:\Program Files\GTK3-Runtime Win64\bin


def unique_filepath(filepath):
    """Generate a unique filename to ensure existing files are not overwritten.

    :param filepath: File-name, including path, to the file location
    :type filepath: str
    :return: Unique filepath (filename and path to file location)
    :rtype: str
    """
    filepath = Path(filepath)
    suffix = filepath.suffix
    name = filepath.name[: -len(filepath.suffix)]
    path = filepath.parent
    count = 0

    while filepath.is_file():
        count += 1
        filepath = Path("{}/{}_{}{}".format(path, name, count, suffix))
    return filepath


def export_file(content, path, overwrite=False):
    """Create a file and export content to it. If overwrite is False a new filename is generated.

    :param content: Text data to be written to file
    :type content: str
    :param path: Path, including filename, for the file to be created
    :type path: str
    :param overwrite: Dictate if an existing file should be overwritten. If False 'path' is amended to be a unique filename, defaults to False
    :type overwrite: bool, optional
    :return: Actual path the file was written to. Note: this may differ from supplied 'path' if 'overwrite' is False)
    :rtype: str
    """
    # convert path to Path object
    path = Path(path)

    if not overwrite:
        path = unique_filepath(path)
    path.touch(exist_ok=True)
    path.write_text(content)
    return path


def load_data(path):
    path = Path(path)
    with path.open() as f:
        return f.read()


# TODO: I think this function is duplicated in core.py - merge?
def update_dict(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d


################################################################
#
# Argparse functions
#
################################################################

# TODO: rename to more generic title - used to load user supplied config too.
def get_diagram_instance(src, instance_name="diagram"):
    src = Path(src)
    spec = importlib.util.spec_from_file_location(f"{src.stem}", src.name)
    user_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(user_module)
    return getattr(user_module, instance_name)


def create_stylesheet(src, path, instance_name="diagram", overwrite=False):
    """Create a stylesheet if none supplied."""
    from pinout import config, style_tools, templates
    from pinout.components.annotation import AnnotationLabel
    from pinout.components.layout import Panel, Diagram_2Columns, Diagram_2Rows
    from pinout.components.legend import Legend
    from pinout.components.pinlabel import PinLabel
    from pinout.components.integrated_circuits import DIP, QFP

    # Save CWD and return to it and end of function
    # incase multiple diagrams are being built from a script
    init_dir = Path.cwd()

    src = Path(src)
    os.chdir(src.parent)
    sys.path.append("")

    diagram = get_diagram_instance(src, instance_name)

    # Extract css class tags from PinLabels
    lbls = diagram.find_children_by_type(diagram, PinLabel)
    tags = list(set([tag for label in lbls for tag in label.tag.strip().split(" ")]))
    if config.pinlabel["tag"] in tags:
        tags.remove(config.pinlabel["tag"])

    context = {}
    if tags:
        context["tags"] = style_tools.assign_color(tags)
        context["pinlabel"] = config.pinlabel
    if diagram.find_children_by_type(diagram, Legend):
        context["legend"] = config.legend
    if diagram.find_children_by_type(diagram, Panel):
        context["panel"] = config.panel
    if diagram.find_children_by_type(diagram, AnnotationLabel):
        context["annotation"] = config.annotation
    if diagram.find_children_by_type(diagram, DIP):
        context["ic_dip"] = config.ic_dip
    if diagram.find_children_by_type(diagram, QFP):
        context["ic_qfp"] = config.ic_qfp
    if isinstance(diagram, Diagram_2Columns) or isinstance(diagram, Diagram_2Rows):
        context["diagram_presets"] = config.diagram_presets

    css_tplt = templates.get("stylesheet.j2")
    css = css_tplt.render(css=context)

    # Return to folder script was launched from
    os.chdir(init_dir)

    # Export stylesheet file
    if not overwrite:
        path = unique_filepath(path)
    with open(path, "w") as f:
        f.write(css)

    print(f"Stylesheet created: '{path}'")
    print(
        f"To use insert the following line into '{src.name}' after '{instance_name}' is declared:"
    )
    print(f"{instance_name}.add_stylesheet('{path}')")


def duplicate(resource_name, *args):
    """Duplicate resources from *pinout* package.

    A variety of resources can be copied via this function, access is available via the command line::

        py -m pinout.manager --duplicate <resource_name>

    Where :code:`<resource_name>` can be any one of the following resources:

        + config
        + quick_start

    :param resource_name: Name of resource to duplicate.
    :type resource_name: string
    """

    resources = {
        "quick_start": [
            ("quick_start", "data.py"),
            ("quick_start", "hardware.png"),
            ("quick_start", "pinout_diagram.py"),
            ("quick_start", "styles.css"),
        ],
        "config": [("config.py",)],
        "kicad": [("pinout_kicad_example.zip",)],
    }

    resource_package = "pinout"
    for path in resources[resource_name]:
        resource_path = "/".join(("resources", *path))
        data = pkg_resources.resource_string(resource_package, resource_path)
        filename = path[-1]
        with open(filename, "wb") as f:
            f.write(data)
        print(f"{filename} duplicated.")


def export_diagram(src, dest, instance_name="diagram", overwrite=False):
    # reset core.diagram_id.counter. Pytest can create multiple diagrams
    # and requires each new diagram's id counter to commence at 0.
    core.diagram_id.counter = 0

    # Save CWD and return to it and end of function
    # incase multiple diagrams are being built from a script
    init_dir = Path.cwd()

    # Save dest for consistent printing at end
    raw_dest = dest

    # Create 'dest' folder(s) and file - file must exist for Path.resolve() to function as expected.
    dest = Path(dest)
    dest.parent.mkdir(parents=True, exist_ok=True)
    if not overwrite:
        dest = unique_filepath(dest)
    dest.touch(exist_ok=True)
    # Convert 'dest' to absolute so CWD can be changed without affecting 'dest'.
    dest = dest.resolve()

    # Change CWD to folder 'src' module is located.
    # This allows pinout.manager to be invoked from any location
    # but keeps paths in the src script relative to that script.
    src = Path(src)
    os.chdir(src.parent)
    sys.path.append("")

    diagram = diagram = get_diagram_instance(src, instance_name)

    # Prepare linked media depending on output
    if dest.suffix == ".svg":
        # Update relative image.src to be relative to destination
        images = diagram.find_children_by_type(diagram, core.Image)
        for img in images:
            # Follow image references
            while isinstance(img.src, core.Image):
                img = img.src

            if not img.embed and not img.src.is_absolute():
                img.src = os.path.relpath(
                    Path.cwd().joinpath(img.src), Path.cwd().joinpath(dest.parent)
                )

        # Update relative stylesheet.src to be relative to destination.
        stylesheets = diagram.find_children_by_type(diagram, core.StyleSheet)
        for css in stylesheets:
            if not css.embed and not css.src.is_absolute():
                css.src = os.path.relpath(
                    Path.cwd().joinpath(css.src), Path.cwd().joinpath(dest.parent)
                )
    else:
        # Embed styles - required for cairosvg to render correctly
        stylesheets = diagram.find_children_by_type(diagram, core.StyleSheet)
        for css in stylesheets:
            css.embed = True

    # Render final SVG file
    try:
        if dest.suffix == ".svg":
            dest.write_text(diagram.render())

        elif dest.suffix == ".png":
            cairosvg.svg2png(bytestring=diagram.render(), write_to=dest.as_posix())

        elif dest.suffix == ".pdf":
            cairosvg.svg2pdf(bytestring=diagram.render(), write_to=dest.as_posix())

        elif dest.suffix == ".ps":
            cairosvg.svg2ps(bytestring=diagram.render(), write_to=dest.as_posix())

        print(f"'{raw_dest}' exported successfully.")

    except Exception as e:
        print(f"The export falied! Error: {e}")

    # Return the CWD to the initial directory
    os.chdir(init_dir)


def create_kicad_lib(dest=".", config_file=None, version=None):
    from datetime import datetime
    from pinout import templates, config

    version = int(version or 6)

    # Load config and create library context
    if version >= 6:
        context = config.kicad_6_footprints
        context["component_type"] = "footprint"
    else:
        context = config.kicad_5_footprints
        context["component_type"] = "module"

    if config_file:
        context = update_dict(
            context, get_diagram_instance(config_file, "kicad_6_footprints")
        )

    context["timestamp"] = datetime.today().strftime("%Y%m%d")

    # Override config version if none supplied
    context["version"] = float(version) or float(context["version"])

    folder = Path(dest, "pinout.pretty")
    folder.mkdir(parents=True, exist_ok=True)
    footprints = ["Annotation", "Origin", "PinLabel"]

    for fp in footprints:
        tplt = templates.get(f"kicad_footprint_lib/{fp}.j2")
        filepath = folder / (fp + ".kicad_mod")
        with filepath.open(mode="w") as f:
            f.write(tplt.render(context))
    print("pinout footprint library for KiCad created successfully.")


def __main__():

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--kicad_lib",
        action="store",
        nargs="+",
        help="Create KiCad footprint library. Example usage: python -m pinout.manager --kicad_lib <destination folder> <optional:config_file>",
    )

    parser.add_argument(
        "-d",
        "--duplicate",
        action="store",
        choices=["quick_start", "config", "kicad"],
        help="duplicate pinout resources",
    )

    parser.add_argument(
        "-e",
        "--export",
        nargs="+",
        action="store",
        help="example usage: python -m pinout.manager -e <module name> <export filename> [<instance name>]\n <instance name> defaults to 'diagram'",
    )

    parser.add_argument(
        "--css",
        nargs="+",
        action="store",
        help="example usage: python -m pinout.manager --css <module name> <export filename> [<instance name>]\n <instance name> defaults to 'diagram'",
    )

    parser.add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        help="Enable file overwriting.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="store",
        help="Accepts integer only. Example usage (with --kicad_lib): python -m pinout.manager --kicad_lib <destination folder> <optional:config_file> -v 6",
    )

    args = parser.parse_args()
    if args.duplicate:
        duplicate(args.duplicate, args.overwrite)

    if args.export:
        export_diagram(*args.export, overwrite=args.overwrite)

    if args.css:
        create_stylesheet(*args.css, overwrite=args.overwrite)

    if args.kicad_lib:
        create_kicad_lib(*args.kicad_lib, version=args.version)


if __name__ == "__main__":
    __main__()



import math
from pinout.core import Group, SvgShape, Rect, BoundingCoords, Coords
from pinout import config
from pinout.components.pinlabel import PinLabelGroup
from pinout.components.leaderline import Curved


class Pin(Group):
    def __init__(self, width, height, polarity_mark=False, **kwargs):
        super().__init__(**kwargs)
        self.polarity_mark = polarity_mark
        self._width = width
        self._height = height

        # Add shape so bounding coords are reported correctly
        self.add(
            SvgShape(
                x=-self._width / 2,
                y=-self._height / 2,
                width=self._width,
                height=self._height,
            )
        )

    def render(self):
        self.add(
            Rect(
                x=-self._width / 2,
                y=-self._height / 2,
                width=self._width,
                height=self._height,
                tag="pin__leg",
            )
        )
        # Add polarity marking
        if self.polarity_mark:
            radius = self.config["radius"]
            self.add(
                Rect(
                    x=-self._width / 2 - radius * 3,
                    y=-radius,
                    width=radius * 2,
                    height=radius * 2,
                    corner_radius=radius,
                    tag=self.config["tag"],
                )
            )
        return super().render()


class DIP(Group):
    """Create a dual in-line package graphic"""

    def __init__(self, pin_count, width, height, **kwargs):
        if pin_count % 2 != 0:
            raise ValueError("pin_count not divisible by 2.")
        super().__init__(**kwargs)
        self.pin_count = pin_count
        self._width = width
        self._height = height
        self.update_config(config.ic_dip)
        self.inset = BoundingCoords(*self.config["inset"])
        self.add_tag(self.config["tag"])
        self.add(SvgShape(width=width, height=height))

    @property
    def pin_pitch(self):
        body_height = self._height - (self.inset.y1 + self.inset.y2)
        pins_per_side = self.pin_count // 2
        return body_height / (pins_per_side + 1)

    @property
    def pitch_coords(self):
        pitch_x = math.cos(math.radians(self.rotate)) * self.pin_pitch
        pitch_y = math.sin(math.radians(self.rotate)) * self.pin_pitch
        return Coords(pitch_x, pitch_y)

    def pin_coords(self, index, rotate=True):
        """Return coordinates of pin by its index."""
        pins_per_side = self.pin_count // 2
        if index <= self.pin_count // 2:
            # lhs header
            x = self.inset.x1 / 2
        else:
            # rhs header
            index = index - self.pin_count // 2
            x = self._width - self.inset.x2 / 2
        y = self.inset.y1 + self.pin_pitch * ((index - 1) % pins_per_side + 1)

        # calculate for rotation
        if rotate:
            rotate = math.radians(self.rotate)
            rx = math.cos(rotate) * x - math.sin(rotate) * y
            ry = math.sin(rotate) * x + math.cos(rotate) * y
            return Coords(rx, ry)

        return Coords(x, y)

    def render(self):
        # Add body
        x1, y1, x2, y2 = self.inset
        self.add(
            Rect(
                x=x1,
                y=y1,
                width=self._width - (x1 + x2),
                height=self._height - (y1 + y2),
                corner_radius=self.config["body"]["corner_radius"],
                tag=self.config["body"]["tag"],
            )
        )
        # Add pin legs
        for i in range(1, self.pin_count + 1):
            if i <= self.pin_count // 2:
                scale = Coords(-1, 1)
            else:
                scale = Coords(1, 1)
            x, y = self.pin_coords(i, False)
            self.add(
                Pin(
                    width=self.inset.x1,
                    height=self.pin_pitch / 2,
                    x=x,
                    y=y,
                    tag=f"ic__leg pin_{i}",
                    scale=scale,
                    polarity_mark=i == 1,
                    config=self.config["polarity_mark"],
                )
            )

        return super().render()


class QFP(Group):
    """Create a quad flat package graphic"""

    def __init__(self, pin_count, length, **kwargs):
        if pin_count % 4 != 0:
            raise ValueError("pin_count not divisible by 4.")
        super().__init__(**kwargs)
        self.pin_count = pin_count
        self.length = length

        self.update_config(config.ic_qfp)
        self.inset = BoundingCoords(*self.config["inset"])
        self.add_tag(self.config["tag"])

    @property
    def pin_pitch(self):
        body_length = self.length - (self.inset.y1 + self.inset.y2)
        pins_per_side = self.pin_count // 4
        return body_length / (pins_per_side + 1)

    @property
    def pitch_coords(self):
        pitch_x = math.cos(math.radians(self.rotate)) * self.pin_pitch
        pitch_y = math.sin(math.radians(self.rotate)) * self.pin_pitch
        return Coords(pitch_x, pitch_y)

    def pin_coords(self, index, rotate=True):
        """Return coordinates of pin by its index."""
        pins_per_side = self.pin_count // 4

        if index <= self.pin_count * 0.25:
            # bottom
            x = self.inset.x1 + self.pin_pitch * (((index - 1) % pins_per_side) + 1)
            y = self.length - self.inset.y2 / 2
        elif index <= self.pin_count * 0.5:
            # rhs
            x = self.length - self.inset.x2 / 2
            y = (
                self.length
                - self.inset.y2
                - self.pin_pitch * (((index - 1) % pins_per_side) + 1)
            )
        elif index <= self.pin_count * 0.75:
            # top
            x = (
                self.length
                - self.inset.x2
                - self.pin_pitch * (((index - 1) % pins_per_side) + 1)
            )
            y = self.inset.y1 / 2
        else:
            # lhs
            x = self.inset.x2 / 2
            y = self.inset.y1 + self.pin_pitch * (((index - 1) % pins_per_side) + 1)

        # calculate for rotation
        if rotate:
            rotate = math.radians(self.rotate)
            rx = math.cos(rotate) * x - math.sin(rotate) * y
            ry = math.sin(rotate) * x + math.cos(rotate) * y
            return Coords(rx, ry)

        return Coords(x, y)

    def render(self):
        # Add body
        x1, y1, x2, y2 = self.inset
        self.add(
            Rect(
                x=x1,
                y=y1,
                width=self.length - (x1 + x2),
                height=self.length - (y1 + y2),
                corner_radius=self.config["body"]["corner_radius"],
                tag=self.config["body"]["tag"],
            )
        )
        # Add pin legs
        for i in range(1, self.pin_count + 1):
            if i <= self.pin_count * 0.25:
                scale = (1, 1)
                rotate = 90
            elif i <= self.pin_count * 0.5:
                scale = (1, 1)
                rotate = 0
            elif i <= self.pin_count * 0.75:
                scale = (1, 1)
                rotate = -90
            else:
                scale = (-1, 1)
                rotate = 0
            x, y = self.pin_coords(i, False)
            self.add(
                Pin(
                    width=self.inset.x1,
                    height=self.pin_pitch / 2,
                    x=x,
                    y=y,
                    tag=f"ic__leg pin_{i}",
                    scale=scale,
                    rotate=rotate,
                    polarity_mark=i == 1,
                    config=self.config["polarity_mark"],
                )
            )

        return super().render()


def labelled_qfn(labels, length=160, label_start=(100, 20), label_pitch=(0, 30)):
    """Generate a QFP graphic with pin-labels applied."""
    graphic = Group()
    ic = graphic.add(
        QFP(
            pin_count=len(labels),
            length=length,
            rotate=45,
        )
    )
    pins_per_side = int(ic.pin_count * 0.25)

    # Side 1
    graphic.add(
        PinLabelGroup(
            x=ic.pin_coords(1).x,
            y=ic.pin_coords(1).y,
            pin_pitch=(ic.pitch_coords),
            label_start=label_start,
            label_pitch=label_pitch,
            scale=(-1, 1),
            labels=labels[:pins_per_side],
            leaderline=Curved(direction="vh"),
        )
    )

    # Side 2
    graphic.add(
        PinLabelGroup(
            x=ic.pin_coords(pins_per_side * 2).x,
            y=ic.pin_coords(pins_per_side * 2).y,
            pin_pitch=(-ic.pitch_coords.y, ic.pitch_coords.x),
            label_start=label_start,
            label_pitch=label_pitch,
            scale=(1, 1),
            labels=labels[int(pins_per_side * 2) - 1 : pins_per_side - 1 : -1],
            leaderline=Curved(direction="vh"),
        )
    )

    # Side 3
    graphic.add(
        PinLabelGroup(
            x=ic.pin_coords(pins_per_side * 2 + 1).x,
            y=ic.pin_coords(pins_per_side * 2 + 1).y,
            pin_pitch=(-ic.pitch_coords.x, -ic.pitch_coords.y),
            label_start=label_start,
            label_pitch=label_pitch,
            scale=(1, -1),
            labels=labels[pins_per_side * 2 : pins_per_side * 3],
            leaderline=Curved(direction="vh"),
        )
    )

    # Side 4
    graphic.add(
        PinLabelGroup(
            x=ic.pin_coords(pins_per_side * 4).x,
            y=ic.pin_coords(pins_per_side * 4).y,
            pin_pitch=(ic.pitch_coords.y, -ic.pitch_coords.x),
            label_start=label_start,
            label_pitch=label_pitch,
            scale=(-1, -1),
            labels=labels[pins_per_side * 4 - 1 : pins_per_side * 3 - 1 : -1],
            leaderline=Curved(direction="vh"),
        )
    )
    return graphic


def labelled_dip(labels, width=100, height=160, label_start_x=100, label_pitch=(0, 30)):
    """Generate a DIP graphic with pin-labels applied."""
    graphic = Group()
    ic = graphic.add(DIP(len(labels), width, height))
    pins_per_side = len(labels) // 2
    label_pitch = Coords(*label_pitch)
    label_start = (
        label_start_x,
        ((pins_per_side - 1) * ic.pin_pitch - (pins_per_side - 1) * label_pitch.y) / 2,
    )
    # LHS Side
    graphic.add(
        PinLabelGroup(
            x=ic.pin_coords(1).x,
            y=ic.pin_coords(1).y,
            pin_pitch=(-ic.pitch_coords.y, ic.pitch_coords.x),
            label_start=label_start,
            label_pitch=label_pitch,
            scale=(-1, 1),
            labels=labels[:pins_per_side],
        )
    )
    # RHS Side
    graphic.add(
        PinLabelGroup(
            x=ic.pin_coords(pins_per_side + 1).x,
            y=ic.pin_coords(pins_per_side + 1).y,
            pin_pitch=(-ic.pitch_coords.y, ic.pitch_coords.x),
            label_start=label_start,
            label_pitch=label_pitch,
            scale=(1, 1),
            labels=labels[-1 : pins_per_side - 1 : -1],
        )
    )

    return graphic



from pinout import core, config
from pinout.components.layout import Group


class TextBlock(Group):
    """Multiline text component."""

    def __init__(self, content, line_height=None, **kwargs):
        # initialise module attrs
        self._scale = core.Coords(*kwargs.pop("scale", (1, 1)))
        self._content = None
        super().__init__(**kwargs)
        self.update_config(config.textblock)
        self.line_height = line_height or self.config["line_height"]
        self.add_tag(self.config["tag"])
        self.content = content

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, content):
        # Convert string to list
        if isinstance(content, str):
            content = [line.strip() for line in content.split("\n")]
        self._content = content

    def render(self):
        self.add_tag(self.config["tag"])
        y = 0
        for text in self.content:
            self.add(
                core.Text(content=text, x=0, y=y, scale=self._scale, **self.config)
            )
            y += self.line_height * self._scale.y

        return super().render()



legend = [
    ("Analog", "analog"),
    ("Communication", "comms"),
    ("Ground", "gnd"),
    ("GPIO", "gpio"),
    ("Touch", "touch"),
    ("Power", "pwr"),
    ("PWM", "pwm"),
]

# Pinlabels

left_header = [
    [
        ("0", "gpio"),
        ("A0", "analog"),
        ("MISO", "comms"),
    ],
    [
        ("1", "gpio"),
        ("MOSI", "comms", {"body": {"x": 92}}),
    ],
    [
        ("2", "gpio"),
        ("A1", "analog"),
        ("SCLK", "comms"),
    ],
]

lower_header = [
    [
        ("3", "gpio"),
        ("PWM", "pwm"),
    ],
    [
        ("4", "gpio"),
        ("A2", "analog"),
        ("TOUCH", "touch"),
    ],
    [
        ("5", "gpio"),
        ("A3", "analog"),
    ],
]

right_header = [
    [
        ("Vcc", "pwr"),
    ],
    [
        ("GND", "gnd"),
    ],
    [
        ("6", "gpio"),
        ("A4", "analog"),
        ("TOUCH", "touch"),
    ],
]


# Text

title = "<tspan class='h1'>Pinout Quick start</tspan>"

description = """Python tool kit to assist with 
documentation of electronic hardware. 
More info at <tspan class='italic strong'>pinout.readthedocs.io</tspan>"""



from pinout import core, config
from pinout.components.layout import Group


class Swatch(Group):
    """Graphical icon for display in LegendEntry"""

    def __init__(self, width=None, height=None, **kwargs):
        super().__init__(**kwargs)
        self.update_config(config.legend["entry"]["swatch"])
        width = width or self.config["width"]
        height = height or self.config["height"]

        # Rect aligned left hand edge, vertically centered around origin.
        shape = self.add(core.Rect(y=-height / 2, width=width, height=height))
        self.add_tag("swatch")
        shape.add_tag("swatch__body")


class LegendEntry(Group):
    """Legend entry comprised of a swatch and single line of text."""

    def __init__(
        self,
        content,
        width=None,
        height=None,
        swatch=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.update_config(config.legend["entry"])
        self.add_tag(self.config["tag"])

        width = width or self.config["width"]
        height = height or self.config["height"]
        swatch = swatch or {}

        if isinstance(swatch, dict):
            swatch = Swatch(**swatch)

        self.add(
            core.SvgShape(
                width=width,
                height=height,
            ),
        )

        swatch.y = height / 2
        swatch.x = (height - swatch.height) / 2
        self.add(swatch)

        self.add(
            core.Text(
                content,
                x=swatch.bounding_coords().x2 + swatch.x,
                y=self.height / 2,
            )
        )


class Legend(Group):
    """Auto generate a legend component"""

    def __init__(
        self,
        data,
        max_height=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.update_config(config.legend)
        self.add_tag(self.config["tag"])

        max_height = max_height or self.config["max_height"]

        entry_x = 0
        entry_y = 0
        for entry in data:

            if type(entry) is tuple:
                content, tag, *args = entry
                attrs = args[0] if len(args) > 0 else {}
                entry = LegendEntry(content, tag=tag, **attrs, scale=self.scale)

            self.add(entry)

            # Position entry in legend
            if max_height and entry_y + entry.height > max_height:
                entry_x = self.width
                entry_y = 0
            entry.x = entry_x
            entry.y = entry_y
            entry_y += entry.height
